# coding: utf8
# author: zmej74
import os, shutil
import xbmcgui, xbmc
import xml.etree.ElementTree as ET
import commands
import zipfile

__tmp__ = "c:/temp" if os.name == "nt" else "/tmp"
#__tmp__ = "d:/temp"
"""
Классы для управления работой эмуляторов
"""


class emu:
    """
    описывает эмулятор и выполняет с ним базовые действия
    """
    def __init__(self):
        self.emu_id = ""
        self.emu_name = ""
        self.emu_version = ""
        self.emu_bin = ""
        self.start_config_path = ""
        self.config_path = ""
        self.start_bin_path = ""
        self.bin_path = ""
        self.log = None
        self.script_start = ""
        self.command = ""
        self.multiconf = False
        self.islogout = True
        self.cwd = ""

    def logout(self, *messages):
        for message in messages:
            if self.islogout:
                if self.log:
                    print >> self.log, message
                else:
                    print message

    def emu_running(self):
        """
        возвращает True если эму запущен
        """
        result = False
        try:
            #output = commands.getoutput("ps -elf --cols 300 | grep %s | grep -v grep | wc -l" % self.emu_bin)
            output = commands.getoutput("pidof %s | wc -w" % self.emu_bin)
            result = False if output == "0" else True
        except:
            result = False
        return result

    def emu_start(self):
        """
        запускает эмулятор
        """
        try:
            #проверяем папку конфига
            if not os.path.isdir(self.config_path):
                os.makedirs(self.config_path, 0644)
                self.logout("Config dir [ %s ] created." % self.config_path )
            if len(os.listdir( self.config_path )) == 0:
                cmd = "cp %s %s" % (os.path.join(self.start_config_path,"*"), self.config_path)
                os.system( cmd )
                self.logout("Copy config from [ %s ] to [ %s ] done." % (self.start_config_path, self.config_path) )
            #проверяем папку бинарника и если нужно- создаем
            if not os.path.isdir(self.bin_path):
                os.makedirs(self.bin_path, 0755)
                self.logout("Bin dir [ %s ] created." % self.bin_path )
            if not os.path.isfile( os.path.join(self.bin_path, self.emu_bin)):
                shutil.copy(os.path.join(self.start_bin_path,self.emu_bin), os.path.join(self.bin_path, self.emu_bin))
                self.logout("Copy binary[ %s ] from [ %s ] to [ %s ] done." % (self.emu_bin, self.start_bin_path, self.bin_path) )
            #проверяем наличие скрипта автозапуска и если нужно - создаем
            if not os.path.isfile(self.script_start):
                sc = open(self.script_start,  "w")
                print >> sc, "#!/bin/sh"
                print >> sc, "if [ -d %s ]; then" % (self.cwd + "/")
                print >> sc, "\t", self.command
                print >> sc, "fi"
                sc.close()
                self.logout("Start script [ %s ] created." % self.script_start)
            os.chmod(self.script_start,  0755)
            self.logout("Atribute [ %s ] changed." % self.script_start)
            os.chmod( os.path.join(self.bin_path,  self.emu_bin ), 0755 )
            self.logout("Atribute [ %s ] changed." % os.path.join(self.bin_path,  self.emu_bin ) )
            #запускаем
            if not self.emu_running():
                os.system(self.script_start)
        except:
            print "Error during start emu %s" % self.emu_bin
            self.showmessage(self.strlanguage(30063, "Error during start emu "+ self.emu_bin), timewait = 5000, icon=xbmcgui.NOTIFICATION_ERROR)
            

    def emu_stop(self):
        """
        останавливает эмулятор - убивает в процессах и удаляет скрипт запуска
        """
        flag = False
        if self.emu_running():
            cmd = "killall -9 %s" % self.emu_bin
            self.logout("Command for stop = %s" % cmd)
            os.system( cmd )
            self.logout("Process [ %s ] stopped." % self.emu_bin)
            flag = True
        # удаляем скрипт автозапуска
        if os.path.isfile(self.script_start):
            os.remove(self.script_start)
            self.logout("Start script [ %s ] deleted." % self.script_start)
        return flag

    def emu_remove(self):
        """
        удаляет эмулятор - убивает в процессах, удаляет скрипт запуска, удаляет бинарник
        """
        self.emu_stop()
        # удаляем бинарник и папку
        fn = os.path.join(self.bin_path, self.emu_bin)
        if os.path.isfile(fn):
            os.remove(fn)
            self.logout("Binary file [ %s ] deleted." % fn)
        if os.path.isdir(self.bin_path) and len(os.listdir( self.bin_path )) == 0:
            os.removedirs(self.bin_path)
            self.logout("Binary dir [ %s ] deleted." % fn)

class emu_manager:
    """
    класс для управления работой эмуляторов - запуск, остановка, удаление
    """
    def __init__(self):
        self.emus = []
        self.islogout = True
        self.log = None
        self.addonname = ""
        self.addonid = ""
        self.addon = None   #xbmcaddon.Addon(id=__addonid__)
        self.cwd = ""
        self.profile = ""
        self.icondir = ""
        self.language = None    #__addon__.getLocalizedString
        self.ecminfo_file = "/tmp/ecm.info" if os.name == "posix" else "d:/tmp/ecm.info"

    def loadEmus(self):
        self.logout("Begin loading emu.")
        XML_FILE = os.path.join(self.cwd,"resources",'emus.xml')
        self.logout("File: "+XML_FILE + " . Is == "+str(os.path.isfile(XML_FILE)))
        try:
            tree = ET.parse(XML_FILE)
        except:
            self.logout("Error reading xml")
        root = tree.getroot()
        self.emus[:] = []
        for child_of_root in root:
            # зачитка параметров эмулятора
            oneemu = emu()
            oneemu.log = self.log
            oneemu.cwd = self.cwd
            oneemu.emu_id = child_of_root.attrib["emu_id"]
            oneemu.emu_name = child_of_root.attrib["emu_name"]
            oneemu.start_bin_path = child_of_root.attrib["start_bin_path"].replace("$ADDON_PATH",self.cwd)
            oneemu.emu_bin = child_of_root.attrib["emu_bin"]
            oneemu.bin_path = child_of_root.attrib["bin_path"].replace("$ADDON_PATH",self.cwd)
            oneemu.start_config_path = child_of_root.attrib["start_config_path"].replace("$ADDON_PATH",self.cwd)
            oneemu.multiconf = True if child_of_root.attrib["multiconfig"] == "1" else False
            oneemu.logout("before read setting...")
            if oneemu.multiconf:
                onestr = self.readsetting("%s_%s" % (oneemu.emu_id, "ConfigPath"))
                if onestr:
                    oneemu.config_path = os.path.dirname( onestr + "/" )
                else:
                    oneemu.config_path = child_of_root.attrib["config_path"].replace("$ADDON_PATH",self.cwd)
                    oneemu.config_path = os.path.dirname( oneemu.config_path + "/" )
                    self.writesetting( "%s_%s" % (oneemu.emu_id, "ConfigPath"), oneemu.config_path)
            else:
                oneemu.config_path = child_of_root.attrib["config_path"].replace("$ADDON_PATH",self.cwd)
                oneemu.config_path = os.path.dirname( oneemu.config_path + "/" )
            """onestr = self.readsetting("%s_%s" % (oneemu.emu_id, "ScriptStart"))
            if onestr:
                oneemu.script_start = onestr
            else:
                oneemu.script_start = child_of_root.attrib["script_start"].replace("$EMU_ID",oneemu.emu_id)
                self.writesetting("%s_%s" % (oneemu.emu_id, "ScriptStart"),oneemu.script_start )
            """
            oneemu.script_start = child_of_root.attrib["script_start"].replace("$EMU_ID",oneemu.emu_id)

            onestr = child_of_root.attrib["command"].replace( "$BIN_NAME", os.path.join( oneemu.bin_path, oneemu.emu_bin) )
            oneemu.command = onestr.replace( "$CONFIG_PATH", oneemu.config_path)
            oneemu.emu_version = child_of_root.attrib["emu_version"]
            self.emus.append( oneemu )
        self.logout("%s Emu descriptions loaded." % len(self.emus))
    def startEmu(self, id_emu):
        for oneemu in self.emus:
            if oneemu.emu_id == id_emu:
                startemu = oneemu
            else:
                if oneemu.emu_stop():
                    xbmc.sleep(3000)
        if not startemu.emu_running():
            startemu.emu_start()
            self.showmessage(self.strlanguage(30002,"Start ")+ " " + startemu.emu_name, icon=self.iconname("start.png"))
            xbmc.sleep(5000)

    def stopEmu(self):
        for oneemu in self.emus:
            if oneemu.emu_stop():
                self.showmessage(self.strlanguage(30004,"Stop ") + " "+  oneemu.emu_name, icon=self.iconname("pause.png"))
                #xbmc.sleep(5000)

    def removeEmu(self):
        if self.showyesno( self.strlanguage(30012,"Remove")+" EMU", self.strlanguage(30013,"Delete EMU's binary files?"), self.strlanguage(30014,"(configs don't delete)")):
            self.showmessage( self.strlanguage(30016, "Removing EMU ") + " ", icon=self.iconname("remove.png"))
            for oneemu in self.emus:
                oneemu.emu_remove()
            #xbmc.sleep(5000)


    def infoEmu(self):
        running = None
        for oneemu in self.emus:
            if oneemu.emu_running(): running = oneemu
        if running:
             #Addon
            info = self.strlanguage(30006,"EMU") + ": %s \n" % running.emu_name
            #Config path
            info += self.strlanguage(30008,"Config path") + ": %s \n" % running.config_path
            #Status
            info += self.strlanguage(30009,"Status") + ": " + self.strlanguage(30010,"Is started!")
        else:
            info = self.strlanguage(30009,"Status") + ": " + self.strlanguage(30011,"Does not work...")
        self.showOk( self.strlanguage(30005,"EMU Info"), info )

    def infoEcm(self):
        """
        вывод диалогового окна с информацией о ecm
        """
        fname = self.ecminfo_file
        infoecm = {}
        for oneemu in self.emus:
            if oneemu.emu_running():
                sinfo = [""] * 4
                if os.path.isfile(fname):
                    try:
                        lines = [line.rstrip() for line in open(fname, "r")]
                        if lines[0][0:2] == "==" and lines[0][-2:] == "==":
                            infoecm = self.EI_wiccard(lines)
                        else:
                            infoecm = self.EI_oscam(lines)
                        #sp = " "*3
                        et = infoecm.get("ecm time","").replace("0.","")
                        et = et.replace(".","")
                        infoecm["ecm time"] = et
                        for (k,v) in infoecm.items():
                            if v[:2] == "0x": infoecm[k] = v[2:]
                        #sinfo[0] = "CAID: " + infoecm.get("caid","") + sp + "IDENT: "+infoecm.get("prov","")+ sp + "PID: "+ infoecm.get("pid","")+ sp + "CHID: " + infoecm.get("chid","")
                        #sinfo[1] = "READER: " + infoecm.get("reader","") + sp + "PROTOCOL: "+infoecm.get("protocol","")
                        #sinfo[2] = "SERVER: " + infoecm.get("from","") + sp + "PORT: "+infoecm.get("port","") + sp + ("TIME: %s ms" % et)
                        #sinfo[3] = "\nCW0:" + infoecm.get("cw0","") + "  " + "CW1:"+infoecm.get("cw1","")
                        #self.showOk( oneemu.emu_name, sinfo[0],sinfo[1],sinfo[2]+sinfo[3])
                    except:
                        pass
                else:
                    #self.showOk( oneemu.emu_name, sinfo[0],sinfo[1],sinfo[2]+sinfo[3])
                    pass
                break
        return infoecm
    def EI_oscam(self,lines):
        """
        парсинг содержимого файла ecm.info эмуляторов Oscam
        """
        info = {}
        try:
            for line in lines:
                s = line.split(":")
                info[s[0].strip()] = s[1].strip()
        except:
            pass
        return info

    def EI_wiccard(self,lines):
        """
        парсинг содержимого файла ecm.info эмуляторов Wicardd/mgcamd
        """
        info = {}
        try:
            info["caid"] = lines[0].split(" on ")[1].split(", ")[0].split("CaID ")[1]
            info["pid"] = lines[0].split("pid ")[1].split(" ")[0]
            info["prov"] = lines[1].split(": ")[1]
            info["cw0"] = lines[2].split(": ")[1].replace(" ","")
            info["cw1"] = lines[3].split(": ")[1].replace(" ","")
            s = lines[5][lines[5].index("(")+1:lines[5].index(")")]
            if " at " in s:
                info["protocol"] = s.split(" at ")[0]
                info["from"] = s.split(" at ")[1].split(":")[0]
                info["port"] = s.split(" at ")[1].split(":")[1]
            else:
                info["from"] = s
            info["ecm time"] = lines[6].split(" ")[0]
        except:
            pass
        return info

    def logout(self, *messages):
        if self.islogout:
            for message in messages:
                if self.log:
                    print >> self.log, message
                else:
                    print message

    def iconname(self,fname):
        """
        возвращает полное имя файла иконки
        """
        return os.path.join(self.cwd, "resources","images", fname)

    def showmessage(self, message, timewait = 5000, icon=""):
        if not icon: icon = self.iconname("message.png")
        xbmc.executebuiltin("Notification(%s, %s, %d, %s)" % (self.addonname, message, timewait, icon ) )

    def showOk(self, header, str1, str2="", str3=""):
        xbmcgui.Dialog().ok(header, str1, str2, str3)

    def showyesno(self, header, str1, str2="", str3=""):
        return xbmcgui.Dialog().yesno(header, str1, str2, str3)

    def readsetting(self, parname):
        """
        читает параметры аддона
        """
        return self.addon.getSetting(parname)

    def writesetting(self, parname, value):
        """
        сохраняет параметры аддона
        """
        self.addon.setSetting(parname, value)

    def strlanguage(self, id_str, altstr=""):
        #self.logout("lang= %s, id_str= %s, altstr= %s"%(xbmc.getLanguage(),id_str,altstr))
        onestr = ""
        try:
            #onestr = self.addon.getLocalizedString( id=id_str )
            onestr = self.addon.getLocalizedString( id_str )
        except:
            onestr = ""
        #self.logout("onestr= %s, id_str= %s, altstr= %s"%(onestr,id_str,altstr))
        if onestr:
            return onestr
        else:
            return altstr

    def emufromzip(self,importfile=""):
        """
        зачитывает новые эмуляторы
        """
        flag = False
        if not importfile:
            importfile = xbmcgui.Dialog().browse(1, self.strlanguage(30028,"Zip file"), "files", ".zip")
        if not importfile:
            return flag
        ff = zipfile.ZipFile(importfile)
        info = {}
        error_message = self.strlanguage(30029,"Unknown error!")
        if self.validateZip(ff):
            info = self.readInfo(ff)
            res = self.validateInfo(info,ff)
            if res[0]:
                self.addOneFromZip(info,ff)
                self.writeNewEmuToXML(info)
                error_message = "Ok!"
                flag = True
            else:
                error_message = res[1]
        else:
            error_message = self.strlanguage(30030,"Wrong zip file!")
        self.showOk(" ", error_message)
        ff.close()
        return flag

    def validateZip(self, zfile):
        """
        проверяет архив с импортируемым эмулятором
        """
        #
        flag = True
        ol = zfile.namelist()
        flags = {"info":False, "bin":False, "config":False, "binfile":False, "configfile":False}
        # проверка содержимого
        for one in ol:
            lone = one.split("/")
            if one == "emu.info":
                flags["info"] = True
            elif len(lone) == 2 and lone[0] == "bin" and lone[1] == "":
                flags["bin"] = True
            elif len(lone) == 2 and lone[0] == "config" and lone[1] == "":
                flags["config"] = True
            elif len(lone) == 2 and lone[0] == "bin" and lone[1]:
                flags["binfile"] = True
            elif len(lone) == 2 and lone[0] == "config" and lone[1]:
                flags["configfile"] = True
        for one in flags.values():
            flag = flag and one
        return flag
        #if flags["info"]:
        #    ff.extract("emu.info", extract_cwd)

    def readInfo(self, zfile):
        """
        читает файл с параметрами загружаемого эмулятора
        """
        extract_cwd = __tmp__
        zfile.extract("emu.info", extract_cwd)
        efile = open(os.path.join(extract_cwd,"emu.info"), "r")
        #info = dict.fromkeys(["id","name","bin","config_path","command","version","multiconfig"],"")
        info = dict.fromkeys(["id","name","bin","command","version","multiconfig","binpath","configpath"],"")
        for one in efile.xreadlines():
            s = one.split("#")[0].strip()
            if s:
                para = s.split("=")
                if len(para) == 2 and info.has_key(para[0].strip()):
                    info[para[0].strip()] = para[1].strip()
        return info

    def validateInfo(self, info, zfile):
        """
        проверяет значения emu.info
        """
        #1 Заполнение всех значений
        flag = True
        notemptyval = ["id","name","bin","command","version","multiconfig"]
        for onekey,oneval in info.items():
            if onekey in notemptyval:
                flag = flag and oneval
        if not flag:
            return (False, self.strlanguage(30031,"Wrong emu.info!"))
        #2 правильность id
        if not info["id"].replace("_","").isalnum():
            return (False, self.strlanguage(30032,"Wrong id value!"))
        #4 уникальность id
        flag = True
        for one in self.emus:
            flag = flag and ( info["id"] != one.emu_id )
        if not flag:
            return (False, self.strlanguage(30033,"Duplication of the id's!"))
        #5 уникальность названия
        flag = True
        for one in self.emus:
            flag = flag and ( info["name"] != one.emu_name )
        if not flag:
            return (False, self.strlanguage(30034,"Duplication of the names!"))
        # 6 Уникальность бинарника
        flag = True
        for one in self.emus:
            flag = flag and ( info["bin"] != one.emu_bin )
        if not flag:
            return (False, self.strlanguage(30035,"Duplication of the binary files!"))
        # 7 наличие именно єтого бинарника
        ol = zfile.namelist()
        if not "bin/"+info["bin"] in ol:
            return (False, self.strlanguage(30036,"Wrong binary file!"))
        # 8 если есть параметр binpath проверяем приблизительную правильность
        strval = info.get("binpath","")
        if strval and not (strval[0] == "/" and strval[-1] != "/"):
            return (False, self.strlanguage(30059,"Wrong binary path!"))
        strval = info.get("configpath","")
        if strval and not (strval[0] == "/" and strval[-1] != "/"):
            return (False, self.strlanguage(30060,"Wrong config path!"))
        return (True, "Ok!")

    def addOneFromZip(self, info, zfile):
        """
        распаковывает архив и добавляет
        """
        try:
            # бинарник
            dest_dir = os.path.join(self.cwd,"resources","cams",info["id"])
            binzip = "bin/"+info["bin"]
            #self.showOk(" ",binzip,dest_dir)
            zfile.extract(binzip,dest_dir)
            dest_dir = os.path.join(self.cwd,"resources","cams",info["id"])
            config_files = [ one for one in zfile.namelist() if len(one) > len("config/") and one[:7]=="config/"]
            for one in config_files: zfile.extract(one,dest_dir)
        except:
            self.showOk(self.strlanguage(30025,"Error"), self.strlanguage(30037,"Unknown error in extract!"))

    def writeNewEmuToXML(self, info):
        """
        Дописывает данные нового эмулятора в файл
        """
        xmlname = os.path.join(self.cwd,"resources","emus.xml")
        stradd = '\t<emu emu_id=\"%s\" emu_name=\"%s\" start_bin_path=\"%s\" emu_bin=\"%s\" bin_path=\"%s\" start_config_path=\"%s\" config_path=\"%s\" script_start=\"%s\" command=\"%s\" emu_version=\"%s\" multiconfig=\"%s\"/>\n'
        #	<emu emu_id="wicardd" emu_name="Wicardd 1.19" start_bin_path="$ADDON_PATH/resources/cams/wicardd/bin" emu_bin="wicardd" bin_path="/config/cams/bin" start_config_path="$ADDON_PATH/resources/cams/wicardd/config" config_path="/config/cams/config/wicardd" script_start="/config/post-maruapp/S50_$EMU_ID_plugin.sh" command="$BIN_NAME -d -c $CONFIG_PATH/wicardd.conf" emu_version="1.19" multiconfig="1"/>
        c_path = info["configpath"] if (info.has_key("configpath") and info["configpath"] ) else "/config/cams/config/%s" % info["id"]
        b_path = info["binpath"] if (info.has_key("binpath") and info["binpath"] ) else "/config/cams/bin"
        stradd = stradd % ( info["id"], info["name"], "$ADDON_PATH/resources/cams/%s/bin" % info["id"], info["bin"], b_path,
                            "$ADDON_PATH/resources/cams/%s/config" % info["id"], c_path, "/config/post-maruapp/S50_%s_plugin.sh"  % info["id"],
                            info["command"].replace("&","&amp;"), info["version"], info["multiconfig"] )
        self.writesetting( "%s_%s" % (info["id"], "ConfigPath"), c_path)
        fxml = open(xmlname)
        #self.showOk("Message", "Before read emua.xml!")
        xmlstr = [one for one in fxml.readlines()]
        #self.showOk("Message", "Before write emua.xml!")
        fxml.close()
        fxml = open(xmlname,"w")
        for one in xmlstr:
            if "</emus>" in one:
                fxml.write(stradd)
            fxml.write(one)
        fxml.close()
        self.loadEmus()

    def deleteOneEMU(self, pos):
        """
        удаляет эмулятор из описания и удаляет файлы эмулятора из папки аддона
        """
        """for i in range(len(self.emus)):
            if self.emus[i].emu_id == emu_id: num = i
        else:
            num = -1"""
        num = pos
        emu_id = self.emus[num].emu_id
        dirname = os.path.join(self.cwd,"resources", "cams", emu_id)
        if len(self.emus) <= 1  or ( os.name=="posix" and self.emus[num].emu_running() ):
            return False
        search = 'emu_id="%s"' % emu_id
        xmlname = os.path.join(self.cwd,"resources","emus.xml")
        xmlstr = [one for one in open(xmlname,"r").readlines()]
        fxml = open(xmlname,"w")
        for one in xmlstr:
            if not search in one:
                fxml.write(one)
        fxml.close()
        self.emus[num].emu_remove()
        # 1.1.3
        parname = "%s_%s" % (emu_id, "ConfigPath")
        onestr = self.readsetting( parname )
        if onestr: 
            self.writesetting( parname, '' )
        #
        try:
            shutil.rmtree(dirname)
        except:
            self.logout("Error for shutil.rmtree")

        self.loadEmus()
        return True

    def changeConfigPath(self, num):
        """
        изменяет в настройках путь к папке с конфигом ємулятора
        """
        if self.emus[num].multiconf:
            newpath = xbmcgui.Dialog().browse(0, self.strlanguage(30038,"New config path for")+" "+self.emus[num].emu_name, "files", "", False, False, self.emus[num].config_path)
            #newpath = xbmcgui.Dialog().browse(0, "New config path for "+self.emus[num].emu_name, "files", "")
            if newpath:
                self.emus[num].config_path = newpath
                self.writesetting( "%s_%s" % (self.emus[num].emu_id, "ConfigPath"), self.emus[num].config_path)
